#include <stdio.h>
#include <dos.h>
#include <conio.h>
#include <ctype.h>
#include <stdlib.h>

/* options */

long symtime	= 0;
long microsec	= 50;	/* Sym steps/10 microseconds */
int ENAena		= 0;	/* Flag: generate ENA */
long StartHold	= 40;	/* default start-hold = 4 microseconds */
char * comment = "";

/* simulated IC lines */

int	SDA = 1;
int	SCL = 1;
int ENA = 0;
int SDAIN = 0;	/* for ack */

int I2CDIR = 0; /* direction bit */

void step(long dly)
{
	long t;
	t = dly*microsec/100;
	if (t==0)
		t++;	/* make minimum of 1 simulator step */
	symtime += t;
}

void plot(void)
{
	if(ENAena)
	{
		printf("%5ld\t%d\t%d\t%d\t\t! %s %s %s %s\n"
			,symtime
			,ENA
			,SCL
			,SDA&1
			,(ENA&1)?" |":"| "
			,(SCL&1)?" |":"| "
			,(SDA&1)?" |":"| "
			,comment);
	} else
	{
		printf("%5ld\t%d\t%d\t\t! %s %s %s\n"
			,symtime
			,SCL
			,SDA&1
			,(SCL&1)?" |":"| "
			,(SDA&1)?" |":"| "
			,comment);
	}
	comment="";
}

int datahi(void)
{
	SDA = 1;
	plot();
	return(SDAIN);
}

int datahi1(void)
{
	comment = "1";
	return(datahi());
}

int dataack(void)
{
	SDA = I2CDIR?0:1;
	SDA |= 2;
	comment = ((SDA&1)	? "ACK: H(chip should reply)"
						: "ACK: L(response from master)");
	if(I2CDIR>1)
		comment = "ACK: not valid at end";
	plot();
	return(SDAIN);
}

int datalow(void)
{
	SDA = 0;
	plot();
	return(SDAIN);
}
int datalow1(void)
{
	comment = "0";
	return(datalow());
}

int clockhi(void)
{
	SCL = 1;
/* 	while (SCLIN != 1) {}; */
	plot();
	return(SDAIN);
}

int clocklow(void)
{
	SCL = 0;
	plot();
	return(SDAIN);
}

int i2c_start(void)
{
	int r;
    step(10); /* microseconds * 10 */
	datahi();
    step(5); /* microseconds * 10 */
	clockhi();
	comment="START";
    step(30); /* microseconds * 10 */
	r=datalow();
	step(StartHold); /* microseconds * 10 */
	clocklow();
    step(10); /* microseconds * 10 */
	return (r);
}
int i2c_stop(void)
{
	datalow();
		step(30); /* microseconds * 10 */
	clockhi();
		step(49); /* microseconds * 10 */
	comment="STOP";
	datahi();
		step(49); /* microseconds * 10 */
	return (clockhi()); /* just checks for data released (clk already hi) */
}

int i2c_out(int d)
{
	int r;
	d = (d<<1)|1;	/* add ack bit to byte */
	while ( d ) /* do until no bits left */
	{
		step(25); /* microseconds * 10 */

		(d & 0x100) ? /* select on next data bit */
		((d&0xFF)?
			datahi1():
			dataack()) : /* 1 bit */
		    datalow1() ;  /* 0 bit */

		step(25); /* microseconds * 10 */
		if((d&0xFF)||(I2CDIR!=2))
		{
			r=clockhi();
			step(42); /* microseconds * 10 */
			clocklow(); /* clock the bit */
		}

		d = (d<<1)&0x1fe; /* shift bits until ack 'bit' is sent */

	}
	step(42); /* microseconds * 10 */
	return(r);
}

/*********************************************************************
* getbyte() gets byte from file
* if a comment starting with "#" is found, it will be copied to output
* as a comment starting with "!"
*  RETURNS:
*		EOX if end of xfer
*		EOM if end of message
*
**********************************************************************/

#define EOX -1
#define EOM -2

int getbyte(FILE * fp)
{
	int c,val=0;

	c=getc(fp);
	/* skip whitespace */
	while(isspace(c)&&(c!='\n'))
	{
		c=getc(fp);
	}
	/* transfer comment */

	if(c=='#')
	{
		putchar('!');
		putchar(' ');
		while(((c=getc(fp))!='\n')&& c!=EOF )
			putchar(c);
		putchar('\n');
   		while(isspace(c)) /* skip whitespace and blank lines! */
		{
			c=getc(fp);
		}
	}


	if (c==EOF)
		return(EOX);

	if(isxdigit(c))
	{
		c=toupper(c);
		val<<=4;
		if(c<'A')
		{
			val+=(c-(int)'0');
		}
		else
		{
			val+=(c-(int)'A'+10);
		}
		c=getc(fp);
	}
	else
		return(EOM); /* non-hex character = end of message */

	if(isxdigit(c))
	{
		c=toupper(c);
		val<<=4;
		if(c<'A')
		{
			val+=(c-(int)'0');
		}
		else
		{
			val+=(c-(int)'A'+10);
		}
	}
	else
	{
		ungetc(c,fp);
		return(val);
	}

	return(val);

}

void i2c_block(FILE * fp)
{
	int byte=0,nextbyte=0,rreq,count;

	while(byte!=EOX)
	{
		byte=getbyte(fp);
		if ((byte==EOX)||(byte==EOM))
			continue;
		i2c_start();
		count = 0;
		rreq = (byte&1); /* set to 1 if read request (lsb of slave address) */

		printf("! Slave Address = %02x\n",byte);
		printf("! R/W = %s\n",rreq ? "Read" : "Write");

		I2CDIR=0;		/* first byte = out */

		while((byte!=EOX) && (byte != EOM))
		{
			nextbyte = getbyte(fp); /* look ahead at next byte */
			printf("! Byte #%d = %02x ---%s\n"
					,count
					,byte
					,count==0 ? " (Slave address) ---"
							  : "--------------------");

			if(I2CDIR && ((nextbyte==EOX)||(nextbyte==EOM)))
			{
					I2CDIR=2;	/* last byte of read */
					printf("! (last byte of read)\n");
			}

			if ((count==0) && (rreq==0))
					printf("! Subaddress    = %02x\n",nextbyte);

			i2c_out(byte);		/* output next byte */

			byte=nextbyte;

			count++;
			I2CDIR=rreq;
		}
		i2c_stop();	/* send STOP */
	}
}


main(int argc, char **argv)
{
	FILE * fp =stdin;
	int parm=0;
   long delay=30L,Ehold=150L;
	char * opt;

	while(++parm < argc)
	{
		opt=argv[parm];
		if (*opt == '-')
		{
			opt++;
			while(*opt)
			{
				switch(*opt)
				{
				case 'f':
				case 'F':
					opt++;
					microsec = atol(opt)*10/4; /* calc new steps/microsec */
					opt="";
					break;
				case 'd':
				case 'D':
					opt++;
					delay = atol(opt);   /* delay to first i2c word */
					opt="";
					break;
				case 's':
				case 'S':
					opt++;
					StartHold = atol(opt);   /* Start Hold */
					opt="";
					break;
				case 'e':
				case 'E':
					opt++;
					ENAena++;
					if(isdigit(*opt))
					{
						Ehold = atol(opt);		/* ENA hold */
						opt="";
					}
					break;
				default:
					fprintf(stderr,
"\nIC Simulation File generator for PICSIM\n\
Usage:\n\
%s [-f<freq>] [-d<delay>] [-h<hold>] [-e][<ehold>]] [<file>]\n\
\n\
Where:\n\
<freq>      = PIC clock freq in integer MHz             (default=20)\n\
<delay>     = Delay before start in s/10 (reset delay) (default=15) \n\
<hold>      = Start Hold time in s/10                  (default=40)\n\
-e          = add ENA line to simulation\n\
<ehold>     = ENA hold before START s/10               (default=150)\n\
<file>      = optional I2C data file in format:\n\
				#comment (optional)\n\
				hex bytes of message, starting with address\n\
			  if no file is given, stdin is used.\n\
",argv[0]);
				exit(2);
				}
			}
		}
		else
		if ((fp=fopen(opt,"r"))==(FILE *)NULL)
		{
			fprintf(stderr,"%s: Can't open '%s'",argv[0],opt);
			exit(1);
		}
	}

	printf("! I2C Simulation file generated by %s\n",argv[0]);
	printf("! %d.%d steps/microsecond\n", microsec / 10, microsec % 10);
	printf("!%s\tSCL\tSDA\n",ENAena?"\tENA":"");
	printf("\n");
	printf("STEP%s\tRA1\tRA0\t\t!%s CL DA input pins\n",ENAena?"\tRA2":"",ENAena?" EN":"");

	comment="Initial state";
	plot();

   step(delay); /* convert steps to microseconds * 10 */

/*  address CHIP for write  */

   ENA = 1;       /* enable the card */
	plot();
   if(ENAena)
   {
		step(Ehold);
		plot();
   }

	i2c_block(fp);

   ENA = 0;       /* disable the card */
   if(ENAena)
   {
		step(5);
		plot();
		step(100);
		plot();
   }

   printf("! Done.\n");


return 0;
}
